import signal
import copy
import tkinter as tk
import pybirdsreynolds.const as const
import pybirdsreynolds.params as params
import pybirdsreynolds.variables as variables
import pybirdsreynolds.draw as draw
from tkinter import font
from functools import partial
from pybirdsreynolds.args import compute_args
from pybirdsreynolds.draw import (
    draw_status_overlays,
    update,
    on_resize,
    patch_1,
    patch_2,
)
from pybirdsreynolds.controls import (
    on_other_key,
    start_repeat,
    stop_repeat,
    toggle_pause,
    on_shift_press,
    on_shift_release,
    signal_handler,
    restore_options
)


def app():

    # Set params values: use user-provided options if available,
    # otherwise fall back to *_DEFAULT values.
    options = compute_args()
    for var_name in dir(params):
        if var_name.endswith("_DEFAULT"):
            option_name = var_name[:-8]
            default_value = getattr(params, var_name)
            value = getattr(options, option_name.lower(), default_value)
            setattr(params, option_name, value)
    if not params.COLOR:
        variables.CANVAS_BG = "black"
        variables.FILL_COLOR = "white"
        variables.OUTLINE_COLOR = "white"
    else:
        variables.CANVAS_BG = "#87CEEB"
        variables.FILL_COLOR = "black"
        variables.OUTLINE_COLOR = "black"

    # Save initial default values to allow restoring params later if needed.
    params.MAX_SPEED_INIT = copy.deepcopy(params.MAX_SPEED)
    params.NEIGHBOR_RADIUS_INIT = copy.deepcopy(params.NEIGHBOR_RADIUS)
    params.NUM_BIRDS_INIT = copy.deepcopy(params.NUM_BIRDS)
    params.WIDTH_INIT = copy.deepcopy(params.WIDTH)
    params.HEIGHT_INIT = copy.deepcopy(params.HEIGHT)
    params.REFRESH_MS_INIT = copy.deepcopy(params.REFRESH_MS)
    params.RANDOM_SPEED_INIT = copy.deepcopy(params.RANDOM_SPEED)
    params.RANDOM_ANGLE_INIT = copy.deepcopy(params.RANDOM_ANGLE)
    params.SEP_WEIGHT_INIT = copy.deepcopy(params.SEP_WEIGHT)
    params.ALIGN_WEIGHT_INIT = copy.deepcopy(params.ALIGN_WEIGHT)
    params.COH_WEIGHT_INIT = copy.deepcopy(params.COH_WEIGHT)
    params.SIZE_INIT = copy.deepcopy(params.SIZE)
    params.FONT_SIZE_INIT = copy.deepcopy(params.FONT_SIZE)
    params.FONT_TYPE_INIT = copy.deepcopy(params.FONT_TYPE)
    params.TRIANGLES_INIT = copy.deepcopy(params.TRIANGLES)
    params.FREE_INIT = copy.deepcopy(params.FREE)
    params.COLOR_INIT = copy.deepcopy(params.COLOR)

    # Save and set the parameters displayed in the UI
    param_docs_ihm = {
        name.removesuffix("_DOC"): value
        for name, value in vars(params).items()
        if name.endswith("_DOC") and getattr(params, f"{name[:-4]}_ACTIVATED") == 2
    }
    params.PARAM_ORDER_IHM = list(param_docs_ihm.keys())

    # set tk app
    draw.root = tk.Tk()
    draw.root.title(f"pybirdsreynolds 🐦")
    draw.root.minsize(
        variables.WIDTH_PARAMS + params.WIDTH_MIN + variables.WIDTH_CONTROLS,
        max(params.HEIGHT, const.HEIGHT_PARAMS_CONTROLS_DEFAULT),
    )
    draw.canvas = tk.Canvas(
        draw.root,
        width=variables.WIDTH_PARAMS + params.WIDTH + variables.WIDTH_CONTROLS,
        height=params.HEIGHT,
        bg=variables.CANVAS_BG,
    )
    draw.canvas.pack(fill="both", expand=True, padx=0, pady=0)

    # Handle fonts (must be done after creating the Tk root window).
    const.FONT_TYPE_LIST = [f for f in const.FONT_TYPE_LIST if f in font.families()]
    if params.FONT_TYPE not in const.FONT_TYPE_LIST:
        params.FONT_TYPE = const.FONT_TYPE_LIST[0]
        params.FONT_TYPE_INIT = copy.deepcopy(params.FONT_TYPE)
    draw_status_overlays()
    update()

    # Handle bindings
    draw.root.bind("p", toggle_pause)
    draw.root.bind("<Key>", on_other_key)
    draw.root.bind_all("<Shift_L>", on_shift_press)
    draw.root.bind_all("<Shift_R>", on_shift_press)
    draw.root.bind_all("<KeyRelease-Shift_L>", on_shift_release)
    draw.root.bind_all("<KeyRelease-Shift_R>", on_shift_release)
    draw.canvas.bind(
        "<Configure>", partial(on_resize, on_other_key, start_repeat, stop_repeat)
    )
    signal.signal(signal.SIGINT, signal_handler)

    # lauch tk app
    draw.root.after(100, patch_1)
    draw.root.after(200, patch_2)
    restore_options()
    draw.root.update()
    draw.root.mainloop()



# args.py
import argparse
import importlib
import pybirdsreynolds.params as params


def get_description() -> str:
    return (
        importlib.resources.files("pybirdsreynolds")
        .joinpath("DESCRIPTION.txt")
        .read_text(encoding="utf-8")
        .strip()
    )


def get_epilog() -> str:
    return (
        importlib.resources.files("pybirdsreynolds")
        .joinpath("EPILOG.txt")
        .read_text(encoding="utf-8")
        .strip()
    )


def display_range(prefix):
    """
    Return a readable description of valid values for a parameter,
    including ranges, free-mode limits, and default value.
    """
    value_default = getattr(params, f"{prefix}_DEFAULT")

    # Case: integer parameter with range constraints
    if not isinstance(value_default, bool) and isinstance(value_default, int):
        value_min = getattr(params, f"{prefix}_MIN")
        value_max = getattr(params, f"{prefix}_MAX")
        value_free_min = getattr(params, f"{prefix}_FREE_MIN")
        value_free_max = getattr(params, f"{prefix}_FREE_MAX")
        parts = []

        if value_min is not None and value_max is not None:
            parts.append(f"between {value_min} and {value_max}")
        elif value_min is not None:
            parts.append(f">= {value_min}")
        elif value_max is not None:
            parts.append(f"<= {value_max}")
        else:
            parts.append(f"with no limit")

        if value_free_min is not None and value_free_max is not None:
            parts.append(f"if --free between {value_free_min} and {value_free_max}")
        elif value_free_min is not None:
            parts.append(f"if --free >= {value_free_min}")
        elif value_free_max is not None:
            parts.append(f"if --free <= {value_free_max}")
        else:
            parts.append(f"if --free with no limit")

        if value_default is not None:
            parts.append(f"default: {value_default}")

        return f"{' , '.join(parts)}" if parts else ""

    # Case: boolean parameter
    elif isinstance(value_default, bool):
        return f"boolean value, default: {value_default}"

    # Case: string (or other types)
    else:
        return f"string value, default: {value_default}"


def check_values(prefix, free, value, parser):
    """
    Validate a parameter value against defined min/max ranges,
    using free-mode limits if `free` is True.
    """
    value_min = getattr(params, f"{prefix}_MIN")
    value_max = getattr(params, f"{prefix}_MAX")
    value_free_min = getattr(params, f"{prefix}_FREE_MIN")
    value_free_max = getattr(params, f"{prefix}_FREE_MAX")

    # Case: no free
    if not free:
        if (
            value_min is not None
            and value_max is not None
            and (value < value_min or value > value_max)
        ):
            parser.error(
                f"{prefix.lower()} must be between {value_min} and {value_max}"
            )
        elif value_min is None and value_max is not None and value > value_max:
            parser.error(f"{prefix.lower()} must <= {value_max}")
        elif value_min is not None and value_max is None and value < value_min:
            parser.error(f"{prefix.lower()} must >= {value_min}")
    # Case: free
    else:
        if (
            value_free_min is not None
            and value_free_max is not None
            and (value < value_free_min or value > value_free_max)
        ):
            parser.error(
                f"{prefix.lower()} must be between {value_free_min} and {value_free_max}"
            )
        elif (
            value_free_min is None
            and value_free_max is not None
            and value > value_free_max
        ):
            parser.error(f"{prefix.lower()} must <= {value_free_max}")
        elif (
            value_free_min is not None
            and value_free_max is None
            and value < value_free_min
        ):
            parser.error(f"{prefix.lower()} must >= {value_free_min}")


def create_parser():
    controls_text = "\n".join(
        f"  [{getattr(params, name.replace('_TEXT', '_COMMAND'))}] : {getattr(params, name)}"
        for name, value in params.__dict__.items()
        if name.endswith("_TEXT") and getattr(params, f"{name[:-5]}_ACTIVATED") > 0
    )

    parser = argparse.ArgumentParser(
        description=get_description() + "\n\n" + f"controls:\n{controls_text}",
        epilog=get_epilog(),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    for name in dir(params):
        if not name.endswith("_DOC"):
            continue

        prefix = name[:-4]
        activated = getattr(params, f"{prefix}_ACTIVATED")
        if activated == 0:
            continue

        default_name = f"{prefix}_DEFAULT"
        if not hasattr(params, default_name):
            continue

        default_value = getattr(params, default_name)
        arg_name = "--" + prefix.lower()
        if isinstance(default_value, bool):
            parser.add_argument(
                arg_name,
                action="store_true",
                default=default_value,
                help=getattr(params, name) + " (" + display_range(prefix) + ")",
            )
        elif isinstance(default_value, int):
            parser.add_argument(
                arg_name,
                type=int,
                default=default_value,
                help=getattr(params, name) + " (" + display_range(prefix) + ")",
            )
        elif isinstance(default_value, str):
            parser.add_argument(
                arg_name,
                type=str,
                default=default_value,
                help=getattr(params, name) + " (" + display_range(prefix) + ")",
            )
    return parser


def compute_args():
    parser = create_parser()
    args = parser.parse_args()

    for name in dir(params):
        if not name.endswith("_DOC"):
            continue

        prefix = name[:-4]
        hide = getattr(params, f"{prefix}_ACTIVATED")
        if hide == 2:
            continue

        default_name = f"{prefix}_DEFAULT"
        if not hasattr(params, default_name):
            continue

        default_value = getattr(params, default_name)
        if not isinstance(default_value, int) or isinstance(default_value, bool):
            continue

        arg_value = getattr(args, prefix.lower(), None)
        check_values(prefix, args.free, arg_value, parser)

    return args


def get_help_text():
    parser = create_parser()
    return parser.format_help()



import pybirdsreynolds.const as const

ARROWS_ACTIVATED = 1

SELECTED_INDEX = 0
PARAM_ORDER_IHM = None

# PARAMS


NUM_BIRDS_ACTIVATED = 2
NUM_BIRDS_DOC = "Number of birds"
NUM_BIRDS_DEFAULT = 500
NUM_BIRDS = 500
NUM_BIRDS_MIN = 1
NUM_BIRDS_MAX = 2000
NUM_BIRDS_FREE_MIN = 1
NUM_BIRDS_FREE_MAX = None

COH_WEIGHT_ACTIVATED = 2
COH_WEIGHT_DOC = "Cohesion weight"
COH_WEIGHT_DEFAULT = 1
COH_WEIGHT = 1
COH_WEIGHT_MIN = 0
COH_WEIGHT_MAX = 10
COH_WEIGHT_FREE_MIN = None
COH_WEIGHT_FREE_MAX = None

ALIGN_WEIGHT_ACTIVATED = 2
ALIGN_WEIGHT_DOC = "Alignment weight"
ALIGN_WEIGHT_DEFAULT = 2
ALIGN_WEIGHT = 2
ALIGN_WEIGHT_MIN = 0
ALIGN_WEIGHT_MAX = 10
ALIGN_WEIGHT_FREE_MIN = None
ALIGN_WEIGHT_FREE_MAX = None

SEP_WEIGHT_ACTIVATED = 2
SEP_WEIGHT_DOC = "Separation weight"
SEP_WEIGHT_DEFAULT = 3
SEP_WEIGHT = 3
SEP_WEIGHT_MIN = 0
SEP_WEIGHT_MAX = 10
SEP_WEIGHT_FREE_MIN = None
SEP_WEIGHT_FREE_MAX = None

NEIGHBOR_RADIUS_ACTIVATED = 2
NEIGHBOR_RADIUS_DOC = "Distance to detect neighbors"
NEIGHBOR_RADIUS_DEFAULT = 50
NEIGHBOR_RADIUS = 50
NEIGHBOR_RADIUS_MIN = 0
NEIGHBOR_RADIUS_MAX = None
NEIGHBOR_RADIUS_FREE_MIN = 0
NEIGHBOR_RADIUS_FREE_MAX = None

WIDTH_ACTIVATED = 2
WIDTH_DOC = "Area width"
WIDTH_DEFAULT = 1000
WIDTH = 1000
WIDTH_MIN = 200
WIDTH_MAX = None
WIDTH_FREE_MIN = 200
WIDTH_FREE_MAX = None

HEIGHT_ACTIVATED = 2
HEIGHT_DOC = "Area height"
HEIGHT_DEFAULT = 500
HEIGHT = 500
HEIGHT_MIN = 500
HEIGHT_MAX = None
HEIGHT_FREE_MIN = 500
HEIGHT_FREE_MAX = None

MAX_SPEED_ACTIVATED = 2
MAX_SPEED_DOC = "Maximum speed of birds"
MAX_SPEED_DEFAULT = 5
MAX_SPEED = 5
MAX_SPEED_MIN = 0
MAX_SPEED_MAX = 100
MAX_SPEED_FREE_MIN = 0
MAX_SPEED_FREE_MAX = 100

RANDOM_SPEED_ACTIVATED = 2
RANDOM_SPEED_DOC = "Random speed variation ratio in percentage"
RANDOM_SPEED_DEFAULT = 3
RANDOM_SPEED = 3
RANDOM_SPEED_MIN = 0
RANDOM_SPEED_MAX = 10
RANDOM_SPEED_FREE_MIN = 0
RANDOM_SPEED_FREE_MAX = None

RANDOM_ANGLE_ACTIVATED = 2
RANDOM_ANGLE_DOC = "Random angle variation in degrees"
RANDOM_ANGLE_DEFAULT = 3
RANDOM_ANGLE = 3
RANDOM_ANGLE_MIN = 0
RANDOM_ANGLE_MAX = 10
RANDOM_ANGLE_FREE_MIN = None
RANDOM_ANGLE_FREE_MAX = None

REFRESH_MS_ACTIVATED = 2
REFRESH_MS_DOC = "Refresh interval in milliseconds"
REFRESH_MS_DEFAULT = 1
REFRESH_MS = 1
REFRESH_MS_MIN = 1
REFRESH_MS_MAX = 100
REFRESH_MS_FREE_MIN = 1
REFRESH_MS_FREE_MAX = None

SIZE_ACTIVATED = 2
SIZE_DOC = "Visual size of birds"
SIZE_DEFAULT = 2
SIZE = 2
SIZE_MIN = 1
SIZE_MAX = 3
SIZE_FREE_MIN = 1
SIZE_FREE_MAX = None

TRIANGLES_ACTIVATED = 2
TRIANGLES_DOC = "Render birds as triangles instead of points"
TRIANGLES_DEFAULT = False
TRIANGLES = False

COLOR_ACTIVATED = 2
COLOR_DOC = "Enable colors"
COLOR_DEFAULT = False
COLOR = False

FONT_TYPE_ACTIVATED = 2
FONT_TYPE_DEFAULT = "Nota Sans Mono"
FONT_TYPE = "Nota Sans Mono"
FONT_TYPE_DOC = "Font"

FONT_SIZE_ACTIVATED = 2
FONT_SIZE_DOC = "Font Size"
FONT_SIZE_DEFAULT = 12
FONT_SIZE = 12
FONT_SIZE_MIN = 11
FONT_SIZE_MAX = 12
FONT_SIZE_FREE_MIN = 11
FONT_SIZE_FREE_MAX = 12

FREE_ACTIVATED = 2
FREE_DOC = "Remove parameter limits"
FREE_DEFAULT = False
FREE = False

# CONTROLS
TOGGLE_START_PAUSE_ACTIVATED = 2
TOGGLE_START_PAUSE_ICON = "⏯"
TOGGLE_START_PAUSE_TEXT = "Toggle start / stop"
TOGGLE_START_PAUSE_COMMAND = "p"

NEXT_FRAME_ACTIVATED = 2
NEXT_FRAME_ICON = "⏭"
NEXT_FRAME_TEXT = "Next frame"
NEXT_FRAME_COMMAND = "n"

REGENERATION_ACTIVATED = 2
REGENERATION_ICON = "🐦"
REGENERATION_COMMAND = "b"
REGENERATION_TEXT = "New generation of birds"

RESET_ACTIVATED = 2
RESET_ICON = "⟳"
RESET_TEXT = "Reset all params"
RESET_COMMAND = "r"

TOGGLE_FPS_ACTIVATED = 2
TOGGLE_FPS_ICON = "⏱"
TOGGLE_FPS_TEXT = "Toggle FPS display"
TOGGLE_FPS_COMMAND = "f"

AVERAGE_ACTIVATED = 2
AVERAGE_ICON = "μ"
AVERAGE_TEXT = "Display average position of birds"
AVERAGE_COMMAND = "a"

TOGGLE_MAXIMIZE_ACTIVATED = 2
TOGGLE_MAXIMIZE_ICON = "🗖"
TOGGLE_MAXIMIZE_TEXT = "Maximize/Minimize"
TOGGLE_MAXIMIZE_COMMAND = "m"

HIDE_ACTIVATED = 2
HIDE_ICON = "👁"
HIDE_TEXT = "Hide Params and Controls"
HIDE_COMMAND = "h"

DOC_ACTIVATED = 2
DOC_ICON = "?"
DOC_TEXT = "Open/Close the documentation"
DOC_COMMAND = "d"

SHIFT_ACTIVATED = 1
SHIFT_ICON = "⇧"
SHIFT_TEXT = "With <LEFT>/<RIGHT> Adjust param by ±10 instead of ±1"
SHIFT_COMMAND = "<SHIFT>"

UP_ACTIVATED = ARROWS_ACTIVATED
UP_ICON = "⬆️"
UP_TEXT = "Navigate params"
UP_COMMAND = "<UP>"

DOWN_ACTIVATED = ARROWS_ACTIVATED
DOWN_ICON = "⬇️"
DOWN_TEXT = "Navigate params"
DOWN_COMMAND = "<DOWN>"

LEFT_ACTIVATED = ARROWS_ACTIVATED
LEFT_ICON = "⬅️"
LEFT_TEXT = "Adjust param (-1)"
LEFT_COMMAND = "<LEFT>"

RIGHT_ACTIVATED = ARROWS_ACTIVATED
RIGHT_ICON = "➡️"
RIGHT_TEXT = "Adjust param (+1)"
RIGHT_COMMAND = "<RIGHT>"
